@@ -6,6 +6,7 @@ import json
 import time
 import itertools
 from multiprocessing import Pool
+from sklearn.model_selection import train_test_split
 from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
 from sklearn.metrics import log_loss, mean_squared_error
 from supervised.algorithms.registry import (
@@ -13,6 +14,7 @@ from supervised.algorithms.registry import (
     MULTICLASS_CLASSIFICATION,
     REGRESSION,
 )
+from supervised.exceptions import AutoMLException
 
 
 def get_binary_score(X_train, y_train, X_test, y_test):
@@ -54,6 +56,7 @@ def get_score(item):
         diff_score = scorer(x_train, y_train, x_test, y_test)
     except Exception as e:
         diff_score = None
+        print(str(e))
 
     try:
         a, b = (
@@ -78,6 +81,7 @@ def get_score(item):
         x_test = np.divide(a, b, out=np.zeros_like(a), where=b != 0).reshape(-1, 1)
         ratio_2_score = scorer(x_train, y_train, x_test, y_test)
     except Exception as e:
+        print(str(e))
         ratio_2_score = None
 
     return (diff_score, ratio_1_score, ratio_2_score)
@@ -96,6 +100,8 @@ class GoldenFeaturesTransformer(object):
         else:
             self._scorer = get_regression_score
 
+        self._error = None
+
         if results_path is not None:
             self._result_file = os.path.join(results_path, "golden_features.json")
             self.try_load()
@@ -103,6 +109,16 @@ class GoldenFeaturesTransformer(object):
     def fit(self, X, y):
         if self._new_features:
             return
+        if self._error is not None and self._error:
+            raise AutoMLException(
+                "Golden Features not created due to error (please check errors.md)."
+            )
+            return
+        if X.shape[1] == 0:
+            self._error = f"Golden Features not created. No continous features. Input data shape: {X.shape}, {y.shape}"
+            self.save(self._result_file)
+            raise AutoMLException("Golden Features not created. No continous features.")
+
         start_time = time.time()
         combinations = itertools.combinations(X.columns, r=2)
         items = [i for i in combinations]
@@ -110,15 +126,7 @@ class GoldenFeaturesTransformer(object):
             si = np.random.choice(len(items), 250000, replace=False)
             items = [items[i] for i in si]
 
-        middle, stop = 2500, 5000
-        if X.shape[0] < 5000:
-            stop = X.shape[0]
-            middle = int(stop / 2)
-
-        X_train = X[:middle]
-        X_test = X[middle:stop]
-        y_train = y[:middle]
-        y_test = y[middle:stop]
+        X_train, X_test, y_train, y_test = self._subsample(X, y)
 
         for i in range(len(items)):
             items[i] += (X_train, y_train, X_test, y_test, self._scorer)
@@ -132,7 +140,9 @@ class GoldenFeaturesTransformer(object):
         #    scores += [get_score(item)]
 
         if not scores:
-            return
+            self._error = f"Golden Features not created. Empty scores. Input data shape: {X.shape}, {y.shape}"
+            self.save(self._result_file)
+            raise AutoMLException("Golden Features not created. Empty scores.")
 
         result = []
         for i in range(len(items)):
@@ -161,6 +171,7 @@ class GoldenFeaturesTransformer(object):
                 ]
             )
             self._new_columns += [new_col]
+            print(f"Add Golden Feature: {new_col}")
 
         self.save(self._result_file)
 
@@ -188,6 +199,8 @@ class GoldenFeaturesTransformer(object):
             "result_file": self._result_file,
             "ml_task": self._ml_task,
         }
+        if self._error is not None and self._error:
+            data_json["error"] = self._error
         return data_json
 
     def from_json(self, data_json):
@@ -195,6 +208,7 @@ class GoldenFeaturesTransformer(object):
         self._new_columns = json.loads(data_json.get("new_columns", []))
         self._result_file = data_json.get("result_file")
         self._ml_task = data_json.get("ml_task")
+        self._error = data_json.get("error")
 
     def save(self, destination_file):
         with open(destination_file, "w") as fout:
@@ -203,3 +217,37 @@ class GoldenFeaturesTransformer(object):
     def try_load(self):
         if os.path.exists(self._result_file):
             self.from_json(json.load(open(self._result_file, "r")))
+
+    def _subsample(self, X, y):
+
+        MAX_SIZE = 5000
+        TRAIN_SIZE = 2500
+
+        shuffle = True
+        stratify = None
+
+        if X.shape[0] > 5000:
+            if self._ml_task != REGRESSION:
+                stratify = y
+            X_train, _, y_train, _ = train_test_split(
+                X, y, train_size=MAX_SIZE, shuffle=shuffle, stratify=stratify
+            )
+            if self._ml_task != REGRESSION:
+                stratify = y_train
+
+            X_train, X_test, y_train, y_test = train_test_split(
+                X_train,
+                y_train,
+                train_size=TRAIN_SIZE,
+                shuffle=shuffle,
+                stratify=stratify,
+            )
+        else:
+            if self._ml_task != REGRESSION:
+                stratify = y
+            train_size = X.shape[0] // 2
+            X_train, X_test, y_train, y_test = train_test_split(
+                X, y, train_size=train_size, shuffle=shuffle, stratify=stratify
+            )
+
+        return X_train, X_test, y_train, y_test
